/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    DynamicReductClassifier.java
 *    Copyright (C) 1999 Eibe Frank, Len Trigg
 *
 */

package weka.classifiers;

import java.util.Enumeration;
import java.util.Vector;

import weka.core.Instances;
import weka.core.Option;
import weka.core.Utils;
import weka.roughset.RelativeReductComputation;
import weka.roughset.Rules;

class DynamicReductClassifier extends RuleClassifier {
  
  static int NUM_SUBTABLES = 10;
  static int DEBUG_LEVELS = 3;
  static double CONFIDENCE = 0.1;
  static double MAX_ERROR = 0.05;
  static double DYNAMIC_STABILITY = 0.8;
  static double REDUCTS_RATIO = 0.2;
  private int debug = 0;
  private int numSubtables = DynamicReductClassifier.NUM_SUBTABLES;
  private double reductsRatio = DynamicReductClassifier.REDUCTS_RATIO;
  private double maxError = DynamicReductClassifier.MAX_ERROR;
  private double confidence = DynamicReductClassifier.CONFIDENCE;
  private double dynamicStability = 
    DynamicReductClassifier.DYNAMIC_STABILITY;
  private Instances instances;

  public DynamicReductClassifier(){}

  /**
   *
   * @param data set of instances serving as training data 
   * @exception Exception if the classifier has not been 
   * generated successfully
   */
  public void buildClassifier(Instances data) {
    
    instances = data;
    try {
      RelativeReductComputation relRed = new RelativeReductComputation(instances);
      rules = new Rules(data);
      rules.addRules(relRed.dynamicReducts(numSubtables,confidence, 
					   maxError,dynamicStability, 
					   (int)(reductsRatio*instances.
						 numInstances())));
    
    if ( debug >= 2 ) {
      printInfo();
    } // end of if ()
    if ( accuracy < 1 ) {
      if ( debug >= 1 ) {
	System.out.println("Approximating rules");
      } // end of if ()      
      rules.approximateRules(accuracy);
    } // end of if ()
    else if ( reducedSetAccuracy < 1 ) {
      if ( debug >= 1 ) {
	System.out.println("Reducing rule sets");
      } // end of if ()      
      rules.reduceRuleSets(reducedSetAccuracy);
    } // end of else
    if ( debug >= 1 ) {
      System.out.println("Generated rule set:\n" + rules.toString());
    } // end of if ()
    } catch (Exception e) {
      e.printStackTrace();
    } // end of try-catch
    
  }
  
  /**
   * Get the value of reductsRatio.
   * @return value of reductsRatio.
   */
  public double getReductsRatio() {
    return reductsRatio;
  }
  
  /**
   * Set the value of reductsRatio. How many reducts, as a ratio to
   * the number of instances in a dataset, should we strive for when
   * computing dynamic reducts?
   * @param v Value to assign to reductsRatio.  */
  public void setReductsRatio(double  v) {
    this.reductsRatio = v;
  }
  
  public String numSubTablesTipText() {
    return "Set the ratio \"number of reducts/number of instances\"";
  }
  
  /**
   * Get the value of numSubtables.
   * @return value of numSubtables.
   */
  public int getNumSubtables() {
    return numSubtables;
  }
  
  /**
   * Set the value of numSubtables.
   * @param v  Value to assign to numSubtables.
   */
  public void setNumSubtables(int  v) {
    this.numSubtables = v;
  }


  public String numSubtablesTipText(){
    return "Set the number of subtables that will be used for " 
      + "dynamic reduct construction";
  }
  
  /**
   * Get the value of confidence.
   * @return value of confidence.
   */
  public double getConfidence() {
    return confidence;
  }
  
  /**
   * Set the value of confidence.
   * @param v  Value to assign to confidence.
   */
  public void setConfidence(double  v) {
    this.confidence = v;
  }

  public String confidenceTipText() {
    return "Set a value between 0 and 1 denoting the uncertainty allowed. " 
      + "(Low value => larger subtables)";
  }
  
  /**
   * Get the value of maxError.
   * @return value of maxError.
   */
  public double getMaxError() {
    return maxError;
  }
  
  /**
   * Set the value of maxError.
   * @param v  Value to assign to maxError.
   */
  public void setMaxError(double  v) {
    this.maxError = v;
  }
  
  public String maxErrorTipText() {
    return "Set the maximum acceptable hypothesis error value " 
      + "(default 0.05)";
  } 
  
  /**
   * Get the value of dynamicStability.
   * @return value of dynamicStability.
   */
  public double getDynamicStability() {
    return dynamicStability;
  }
  
  /**
   * Set the value of dynamicStability.
   * @param v  Value to assign to dynamicStability.
   */
  public void setDynamicStability(double  v) {
    this.dynamicStability = v;
  }
  
  public String dynamicStabilityTipText() {
    return "Set stability coefficient for dynamic reducts";
  } 

  /**
   */

  public Enumeration listOptions() {

    Vector l = new Vector();

    for (Enumeration e = super.listOptions();e.hasMoreElements(); ) {
      l.addElement((Option)e.nextElement());
    } // end of for ()
    l.addElement(new Option("use Debugging information. " 
			    + "Parameter between 0 and " 
			    + DynamicReductClassifier.DEBUG_LEVELS,
			    "D", 1, "-D <val>"));

    l.addElement(new Option("reduct raTio.","-T", 1, 
			    "-T <reductsRatio>"));

    l.addElement(new Option("number of Subtables.","S", 1, 
			    "-S <numSubTables>"));

    l.addElement(new Option("Confidence value between 0 and 1.","C", 1, 
			    "-C <confidence>"));

    l.addElement(new Option("maximum Error value, between 0 and 1.","E", 1, 
			    "-E <maxError>"));

    l.addElement(new Option("dYnamic reduct stabililty coefficient," 
			    + " between 0 and 1.","Y",1,
			    "-Y <dynamicStability>"));

    return l.elements();
  }

  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */

  public void setOptions(String[] options) throws Exception{    

    super.setOptions(options);
    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > DynamicReductClassifier.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    +  DynamicReductClassifier.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()
    else {
      debug = 0;
    } // end of else

    String reductsRatioOption = Utils.getOption('C', options);
    if ( reductsRatioOption.length() > 0) {
      reductsRatio = Double.parseDouble(reductsRatioOption);
      if (reductsRatio  < 0 || reductsRatio > 1) {
	throw new Exception("Illegal reductsRatio value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      reductsRatio = DynamicReductClassifier.REDUCTS_RATIO;
    } // end of else


    String numSubtablesOption = Utils.getOption('S', options);
    if ( numSubtablesOption.length() > 0) {
      numSubtables = Integer.parseInt(numSubtablesOption);
      if (numSubtables  < 0) {
	throw new Exception("Illegal number of subtables. "
			    + "Choose a value larger than 0.");
      } // end of if ()
    } // end of if ()
    else {
      numSubtables = DynamicReductClassifier.NUM_SUBTABLES;
    } // end of else

    String confidenceOption = Utils.getOption('C', options);
    if ( confidenceOption.length() > 0) {
      confidence = Double.parseDouble(confidenceOption);
      if (confidence  < 0 || confidence > 1) {
	throw new Exception("Illegal confidence value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      confidence = DynamicReductClassifier.CONFIDENCE;
    } // end of else

    String maxErrorOption = Utils.getOption('E', options);
    if ( maxErrorOption.length() > 0) {
      maxError = Double.parseDouble(maxErrorOption);
      if (maxError  < 0 || maxError > 1) {
	throw new Exception("Illegal maxError value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      maxError = DynamicReductClassifier.MAX_ERROR;
    } // end of else


    String dynamicStabilityOption = Utils.getOption('C', options);
    if ( dynamicStabilityOption.length() > 0) {
      dynamicStability = Double.parseDouble(dynamicStabilityOption);
      if (dynamicStability  < 0 || dynamicStability > 1) {
	throw new Exception("Illegal dynamicStability value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      dynamicStability = DynamicReductClassifier.DYNAMIC_STABILITY;
    } // end of else

  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] inheritedOptions = super.getOptions();
    String [] options = new String[inheritedOptions.length+10];
    System.arraycopy(inheritedOptions,0,options,0,inheritedOptions.length);
    int current = inheritedOptions.length;

    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()

    if ( !(reductsRatio == DynamicReductClassifier.REDUCTS_RATIO) ) {
      options[current++] = "-T"; options[current++] = "" + reductsRatio;
    } // end of if ()

    if ( !(numSubtables == DynamicReductClassifier.NUM_SUBTABLES) ) {
      options[current++] = "-S"; options[current++] = "" + numSubtables;
    } // end of if ()

    if ( !(confidence == DynamicReductClassifier.CONFIDENCE) ) {
      options[current++] = "-C"; options[current++] = "" + confidence;
    } // end of if ()

    if ( !(maxError == DynamicReductClassifier.MAX_ERROR) ) {
      options[current++] = "-E"; options[current++] = "" + maxError;
    } // end of if ()

    if ( !(dynamicStability == DynamicReductClassifier.DYNAMIC_STABILITY) ) {
      options[current++] = "-Y"; options[current++] = "" + dynamicStability;
    } // end of if ()

    /* Fill out the rest */

    while (current < options.length) {
      options[current++] = "";
    }

    return options;
  }

  public void printInfo() {
    System.out.println(this.toString());
  }

  public String toString() {
    StringBuffer text = new StringBuffer();

    if (instances == null) {
      text.append(": No model built yet.");
    }
    else {
      text.append(":\n" 
		  + "Rulesets classifying " 
		  + instances.classAttribute().name() 
		  + ":\n");
      text.append(rules.toString());
    } // end of else

    return text.toString();
  }
  
  /**
   * Main method for testing this class
   */
  public static void main(String [] argv) {
    try {
      System.out.println(Evaluation.evaluateModel(new DynamicReductClassifier(), 
						  argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
  }
}

