/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    LEM2.java
 *    Copyright (C) 2002 Ola Leifler
 *
 */

package weka.classifiers;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.estimators.*;
import weka.roughset.*;

/** 
 *
 * @author Ola Leifler (olale@ida.liu.se)
 */

public class LEM2 extends RuleClassifier implements OptionHandler, 
						    WeightedInstancesHandler {

  /**
   * The dataset header for the purposes of printing out a semi-intelligible 
   * model 
   */

  protected Instances instances;
  private int debug = 0;
  static int DEBUG_LEVELS = 3;

  public LEM2(){
  }

  /**
   * Generates the classifier. Follows the article of Grymala-Busse '92
   * almost exactly
   *
   * @param instances set of instances serving as training data 
   * @exception Exception if the classifier has not been generated 
   * successfully */
  public void buildClassifier(Instances i) throws Exception {
 
    instances = i;
    rules = new Rules(instances);

    /* Add all decision classes to pairs */

    boolean certain = true;
    rules.addRules(instances,certain,debug);
    rules.addRules(instances,!certain,debug);
    if ( debug >= 2 ) {
      printInfo();
    } // end of if ()
    if ( accuracy < 1 ) {
      if ( debug >= 1 ) {
	System.out.println("Approximating rules");
      } // end of if ()      
      rules.approximateRules(accuracy);
    } // end of if ()
    else if ( reducedSetAccuracy < 1 ) {
      if ( debug >= 1 ) {
	System.out.println("Reducing rule sets");
      } // end of if ()      
      rules.reduceRuleSets(reducedSetAccuracy);
    } // end of else
    if ( debug >= 1 ) {
      System.out.println("Generated rule set:\n" + rules.toString());
    } // end of if ()
  }

  /**
     Handling local variables
   */

  public int getDebug() {
    return debug;
  }

  public void setDebug(int v){
    debug = v;
  } 

  public String debugTipText(){
    return "Set debugging level between 0 and 3";
  }

  /**
     -D Debug <debuglevel>
     -S {s|m|r|d|p|c} Rule strength calculation.
        s = Simple Strength
	m = Maximal Strength
	d = Stability Strength
   */

  public Enumeration listOptions() {

    Vector l = new Vector();

    for (Enumeration e = super.listOptions();e.hasMoreElements(); ) {
      l.addElement((Option)e.nextElement());
    } // end of for ()
    l.addElement(new Option("Set debugging level. (0<=level<=3)",
			    "D", 1, "-D <level>"));
    return l.elements();
  }

  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */

  public void setOptions(String[] options) throws Exception{    

    super.setOptions(options);
    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > LEM2.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    + LEM2.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] inheritedOptions = super.getOptions();
    String [] options = new String[inheritedOptions.length+10];
    System.arraycopy(inheritedOptions,0,options,0,inheritedOptions.length);
    int current = inheritedOptions.length;

    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()

    /* Fill out the rest */

    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  public void printInfo() {
    System.out.println(this.toString());
  }

  /**
   * Returns a description of the classifier.
   *
   * @return a description of the classifier as a string.
   */
  public String toString() {
    
    StringBuffer text = new StringBuffer();

    if (instances == null) {
      text.append(": No model built yet.");
    }
    else {
      text.append(":\n" 
		  + "Rulesets classifying " 
		  + instances.classAttribute().name() 
		  + ":\n");
      text.append(rules.toString());
    } // end of else
    return text.toString();
  }
    /**
   * Main method for testing this class.
   *
   * @param argv the options
   */
  public static void main(String [] args) {
    try {
      System.out.println(Evaluation.evaluateModel(new LEM2(), args));
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}













