/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Voter.java
 *    Copyleft 2002 Ola Leifler
 *
 */

package weka.classifiers;

import weka.core.*;
import java.util.*;

/**
 * This class implements algorithm 3 in <i>"A comparison of dynamic
 * and non-dynamic Rough Set methods for extracting laws from decision
 * tables", Jan G. Bazan, International Symposium on Methodologies for
 * Intelligent Systems 1998</i> which builds rules for each tested
 * object that is recieved for classification and does not produce any
 * rule set in advance.
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.1 */

public class Voter extends Classifier implements OptionHandler {

  private int debug = 0;
  private Instances instances;
  static int DEBUG_LEVELS = 3;
  private double accuracy = 1;

  
  public Voter(){}

  public void buildClassifier(Instances inst) throws Exception{
    if (inst.classAttribute().isNumeric()) {
      throw new Exception("Can only classify " 
			  + "into discrete classes");
    } // end of if ()
    instances = inst;
  }

  public double classifyInstance(Instance inst) throws Exception{
    double[] votes = new double[inst.numClasses()];
    int count;
    Arrays.fill(votes,0);
    Hashtable equalInstances = new Hashtable(instances.numInstances());
    Hashtable currentEqualInstances = new Hashtable(instances.numInstances());
    Attributes equalAttributes;
    if ( debug >= 1 ) {
      System.out.println("\n\nClassifying instance:\n" + inst.toString());
    } // end of if ()
    for (int i=0;i<instances.numInstances();i++) {
      Instance inst1 = instances.instance(i);
      if ( !equalInstances.containsKey(inst1) || 
	   !((Boolean)equalInstances.get(inst1)).booleanValue() ) {
	equalAttributes = Attributes.equalAttributes(inst1,inst);
	if ( !equalAttributes.isEmpty() ) {
	  if ( debug >= 2 ) {
	    System.out.println("\n\n====================\n"
			       + "Found partial match:\n" 
			       + inst1.toString());
	  } // end of if ()
	  currentEqualInstances.clear();
	  count = 0;
	  for (int j=0;j<instances.numInstances();j++) {
	    Instance inst2 = instances.instance(j);
	    if ( equalAttributes.subsetOf(Attributes.
					  equalAttributes(inst2,inst)) ) {
	      count++;
	      if ( inst1.classValue() == inst2.classValue() ) {
		if ( debug >= 2 ) {
		  System.out.println("Found an equally good " 
				     + "or more specific match " 
				     + "with same decision:\n" 
				     + inst2.toString());		  
		} // end of if ()
		currentEqualInstances.put(inst2,Boolean.TRUE);
	      } // end of if ()
	      else {
		if ( debug >= 2 ) {
		  System.out.println("Found an equally good " 
				     + "or more specific match:\n" 
				     + inst2.toString());
		} // end of if ()
	      } // end of else
	    } // end of if ()	    
	  } // end of for ()
	  if ( currentEqualInstances.size()/count >= accuracy ) {
	    if ( debug >= 2 ) {
	      System.out.println("More than " + accuracy*100 
				 + "% of the matches with the same decision" 
				 + "\n--------------------------------------" 
				 + "---------------");
	    } // end of if ()
	    equalInstances.putAll(currentEqualInstances);
	  } // end of if ()
	} // end of if ()	
      } // end of if ()
    } // end of for ()

    double[] numWithDecision = new double[instances.numClasses()];
    Arrays.fill(numWithDecision,0);

    for (int i=0;i<instances.numInstances();i++) {
      Instance inst3 = instances.instance(i);
      votes[(int)inst3.classValue()] += 
	((equalInstances.containsKey(inst3) &&  
	 ((Boolean)equalInstances.get(inst3)).booleanValue()) ? 1 : 0); 
      numWithDecision[(int)inst3.classValue()]++;
    } // end of for ()
    for (int i=0;i<instances.numClasses();i++) {
      votes[i] /= numWithDecision[i];
    } // end of for ()

    double maxVotes=0;
    double bestDecision=0;
    for (int i=0;i<votes.length;i++) {
      if (votes[i]>maxVotes) {
	maxVotes = votes[i];
	bestDecision = (double)i;
      } // end of if ()
    } // end of for ()
    if ( debug >= 1 ) {
      Iterator iter = equalInstances.keySet().iterator();
      System.out.println("\n\nPartially matching instances: ");
      while ( iter.hasNext() ) {
	Instance inst4 = (Instance)iter.next();
	boolean b = ((Boolean)equalInstances.get(inst4)).booleanValue();
	if ( b ) {
	  System.out.println(inst4.toString());
	} // end of if ()
      } // end of while ()
      System.out.println("From which we obtained a decision: " 
			 + inst.classAttribute().name() 
			 + " = "
			 + inst.classAttribute().value((int)bestDecision) 
			 + "\n********************************************");
    } // end of if ()
    return bestDecision;
  }
  
  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }
    
  public String debugTipText() { 
    return "Display debugging information. Choose a value between 0 and " 
      + Voter.DEBUG_LEVELS;
  }

  /**
   * Get the value of accuracy.
   * @return value of accuracy.
   */
  public double getAccuracy() {
    return accuracy;
  }
  
  /**
   * Set the value of accuracy.
   * @param v  Value to assign to accuracy.
   */
  public void setAccuracy(double  v) {
    this.accuracy = v;
  }
  
  public String accuracyTipText() {
    return "Reduce the rules until the accuracy is at least this value";
  }


  public Enumeration listOptions() {

    Vector l = new Vector();

    l.addElement(new Option("\tSet debugging level. (0<=level<=3)",
			    "D", 1, "-D <level>"));
    l.addElement(new Option("\tSet accuracy. (0<=accuracy<=1)",
			    "A", 1, "-A <accuracy>"));
    return l.elements();
  }
  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */

  public void setOptions(String[] options) throws Exception{    

    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > Voter.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    + Voter.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()

    String accuracyOption = Utils.getOption('A', options);
    if ( accuracyOption.length() > 0) {
      accuracy = Double.parseDouble(accuracyOption);
      if (accuracy  < 0 || accuracy > 1 ) {
	throw new Exception("Illegal accuracy level. "
			    + "Choose a value between 0 and 1");
      } // end of if ()
    } // end of if ()
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {
    String [] options = new String [10];
    int current = 0;

    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()

    if ( accuracy < 1 ) {
      options[current++] = "-A"; options[current++] = "" + accuracy;
    } // end of if ()

    /* Fill out the rest */

    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  public String toString() {
    return "Voter created from given instances.";
  }

  public static void main (String[] args) {
    try {
      System.out.println(Evaluation.evaluateModel(new Voter(), args));
    } catch (Exception e) {
      e.printStackTrace();
      System.err.println(e.getMessage());
    }
  } // end of main ()
  
}
