/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * DynamicReductDiscretize
 * Copyleft (C) 2002 Ola Leifler
 *
 */

package weka.filters;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.roughset.*;

/**
 * Discretize a set of instances using dynamic reducts
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.1
 */

public class DynamicReductDiscretize extends Filter implements OptionHandler{

  static int NUM_SUBTABLES = 20;
  static int DEBUG_LEVELS = 3;
  static int FAST = 2;
  static int REDUCT_ITERATIONS = 30;
  static double CONFIDENCE = 0.1;
  static double MAX_ERROR = 0.05;
  static double REDUCTS_RATIO = 0.4;
  private int debug = 0;
  private int numSubtables = DynamicReductDiscretize.NUM_SUBTABLES;

  /**
     the number of iterations we should use for
     trying to obtain sets of proper reducts. 
  */
  private int reductIterations = DynamicReductDiscretize.REDUCT_ITERATIONS;

  /**
     the number of iterations in the binary search for a
     proper subtable size we should perform. An alternative break
     condition to the calculation of the minimal subtable size.
   */
  
  private int maxIterations = 5;

  /**
     the error margin for the reductsRatio
  */
  private double error = .3;

  /**
     fraction of the reducts for the subtables that should also be
     reducts for the whole table.  
  */
  private double reductsRatio = DynamicReductDiscretize.REDUCTS_RATIO;
  private double maxError = DynamicReductDiscretize.MAX_ERROR;
  private double confidence = DynamicReductDiscretize.CONFIDENCE;

  private double minFraction = .7;

  private int fast = 2;

  /** Store the current cutpoints */
  protected double [][] cutPoints = null;

  public DynamicReductDiscretize() {
    
  }
  
  protected void calculateCutPoints() throws Exception{

    /**
       (1) Generate a number of cuts until they form a superreduct of
       the whole table.
    */
    if ( debug > 0) {
      System.out.print("Step 1a (add the best cuts from each attribute)");    
    } // end of if ()

    Instances insts = getInputFormat();
    
    /**
       We do not allow many inconsistencies int the input data
     */
    
    MDCutPoints cuts = new MDCutPoints(insts);
    Instances binaryInstances = cuts.binaryInstances(insts);
    MDCutPoints resultingCuts = new MDCutPoints();
    int i=0;
    for (;i<insts.numAttributes();i++) {
      if ( (i != insts.classIndex()) && insts.attribute(i).isNumeric() ) {
	CutPoint c = cuts.getMaxCut(insts.attribute(i));
	cuts.removeCut(c);
	resultingCuts.addCut(c);
	if ( debug > 0 ) {
	  System.out.print(".");	  
	} // end of if ()
      } // end of if ()      
    } // end of for ()
    i=0;

    if ( debug > 0 ) {
      System.out.println(" done.");
      System.out.print("Step 1b (sequentially add new ones until reduct" 
		       + " is obtained)");      
    } // end of if ()    

    /**

       Add attributes on the *binary* table until semi-minimal reduct is
       achieved. 

     */


    Reduct r = RelativeReductComputation.
      semiMinimalReduct(resultingCuts.toReduct(binaryInstances).
			getAttributes(),binaryInstances);

    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()
    
    
    /**

       Then, convert this set of attributes into the corresponding set
       of cuts for the original table using our very ugly string
       comparison method when matching attributes to cuts...

     */
    
    resultingCuts = new MDCutPoints(insts).getSubset(r.getAttributes());

    if ( debug > 0 ) {
      System.out.println(" done.");      
    } // end of if ()
    
    /**
       (2) Convert the cutpoints into a set of reducts for the binary table.
    */

    if ( debug > 0 ) {
      System.out.print("Step 2 ");      
    } // end of if ()

    
    Reducts reds = RelativeReductComputation.
      someProperReducts(resultingCuts.toReduct(binaryInstances),
			reductIterations);
    if ( reds.size() == 0 ) {
      throw new Exception("No proper reducts found to superreduct " 
			  + resultingCuts.toString());
    } // end of if ()
    
    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()

    int minSize;
    if ( fast > 1 ) {
      /** This should work well in most cases */

      minSize = (int)(binaryInstances.numInstances()*minFraction);

    } // end of if ()
    else {
      /**
	 The most time-consuming part of this entire process..
      */

      minSize = RoughSet.minSubtableSize(binaryInstances,reductsRatio,
					 error,reds,maxIterations);       
    } // end of else
    
    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()

    /**
       Should we calculate the stability in the quick'n'dirty way with
       a fix number of subtables or in the way Bazan claims it should
       be done?  */

    SubtableChooser chooser = null;
    if ( fast > 0 ) {
      chooser = new SubtableChooser(binaryInstances,minSize,numSubtables);
    } // end of if ()
    else {
      chooser = new SubtableChooser(binaryInstances,minSize,
				    confidence,maxError);
    } // end of else
    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()
    double numChoices = 0;
    double [] stabilities = new double[reds.size()];
    Arrays.fill(stabilities,0);

    if ( debug > 0 ) {
      System.out.println(" done.");
    } // end of if ()

    /** 
	(3) Calculate the stabilities of the reducts on the given
	family of subtables */

    if ( debug > 0 ) {
      System.out.print("Step 3 ");      
    } // end of if ()

    while ( chooser.hasNext() ) {
      numChoices++;
      Instances instances = chooser.next();
      int j=0;
      for (Iterator iter=reds.iterator();iter.hasNext();j++) {
	stabilities[j] += (Reduct.
			   isProperReduct(((Reduct)iter.next()).
			   getAttributes(),instances)?1:0);
      } // end of for ()
      if ( debug > 0 ) {
	System.out.print(".");	
      } // end of if ()
    } // end of while ()

    if ( debug > 0 ) {
      System.out.println(" done.");      
    } // end of if ()
    
    /**
       (4) Choose the most stable proper reduct and select it as a set
       of cuts */

    if ( debug > 0 ) {
      System.out.print("Step 4 ");      
    } // end of if ()
    
    for (int j=0;j<stabilities.length;j++) {
      stabilities[j] /= numChoices;	
    } // end of for ()

    if ( debug > 0 ) {
      System.out.print(".");          
    } // end of if ()
    
    int index = Utils.maxIndex(stabilities);
    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()
    Reduct reduct = reds.get(index);
    if ( debug > 0 ) {
      System.out.print(".");      
    } // end of if ()
    MDCutPoints md = reduct.getCuts();
    md.fillNonDiscretized(insts);
    cutPoints = md.toDoubleArray(insts);
    if ( debug > 0 ) {
      System.out.println(". done.");
    } // end of if ()
  }

  /**
   * Input an instance for filtering. Ordinarily the instance is processed
   * and made available for output immediately. Some filters require all
   * instances be read before producing output.
   *
   * @param instance the input instance
   * @return true if the filtered instance may now be
   * collected with output().
   * @exception IllegalStateException if no input format has been defined.
   */
  public boolean input(Instance instance) {

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (m_NewBatch) {
      resetQueue();
      m_NewBatch = false;
    }
    
    if (cutPoints != null) {
      convertInstance(instance);
      return true;
    }
    else {
      bufferInput(instance);
      return false;
    } // end of else
    
  }


  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any instances contained in the object are ignored - only the
   * structure is required).
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the input format can't be set successfully
   */
  public boolean setInputFormat(Instances instanceInfo) throws Exception {
    super.setInputFormat(instanceInfo);
    cutPoints = null;
    if (instanceInfo.classIndex() < 0) {
      throw new UnassignedClassException("Cannot use class-based " 
					 + "discretization: "
					 + "no class assigned to " 
					 + "the dataset");
    }
    if (!instanceInfo.classAttribute().isNominal()) {
      throw new UnsupportedClassTypeException("Supervised discretization " 
					      + "not possible:"
					      + " class is not nominal!");
      }
    
    // If we implement loading cutfiles, then load 
    //them here and set the output format
    return false;

  }

  /**
   * Signifies that this batch of input to the filter is finished. If the 
   * filter requires all instances prior to filtering, output() may now 
   * be called to retrieve the filtered instances.
   *
   * @return true if there are instances pending output
   * @exception IllegalStateException if no input structure has been defined
   */
  public boolean batchFinished() throws Exception{

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (cutPoints == null) {
      this.calculateCutPoints();
      if ( cutPoints != null && debug > 0 ) {
	for (int i=0;i<cutPoints.length;i++) {
	  double [] tmp = cutPoints[i];
	  if ( tmp != null && tmp.length > 0 ) {	  
	    System.out.print("[");
	    for (int j=0;j<tmp.length;j++) {
	      System.out.print(tmp[j]+(j+1<tmp.length?", ":"]"));
	    } // end of for ()
	  } // end of if ()
	  else {
	    System.out.print("[]");
	  } // end of else
	  System.out.println();
	} // end of for ()
	System.out.println("---------------------------------------------");
      } // end of if ()
      
      this.setOutputFormat();
      // If we implement saving cutfiles, save the cuts here

      // Convert pending input instances
      for(int i = 0; i < getInputFormat().numInstances(); i++) {
	convertInstance(getInputFormat().instance(i));
      }
    } 
    flushInput();

    m_NewBatch = true;
    return (numPendingOutput() != 0);
  }

  /**
   * Set the output format. Takes the currently defined cutpoints and 
   * m_InputFormat and calls setOutputFormat(Instances) appropriately.
   */
  protected void setOutputFormat() {

    if (cutPoints == null) {
      setOutputFormat(null);
      return;
    }
    FastVector attributes = new FastVector(getInputFormat().numAttributes());
    int classIndex = getInputFormat().classIndex();
    for(int i = 0; i < getInputFormat().numAttributes(); i++) {
      if (getInputFormat().attribute(i).isNumeric()) {
		  FastVector attribValues = new FastVector(1);
	  if (cutPoints[i] == null) {
	    attribValues.addElement("All");
	  } else {
	    for(int j = 0; j <= cutPoints[i].length; j++) {
	      if (j == 0) {
		attribValues.addElement("(-inf-"
			+ Utils.doubleToString(cutPoints[i][j], 6) + "]");
	      } else if (j == cutPoints[i].length) {
		attribValues.addElement("("
			+ Utils.doubleToString(cutPoints[i][j - 1], 6) 
					+ "-inf)");
	      } else {
		attribValues.addElement("("
			+ Utils.doubleToString(cutPoints[i][j - 1], 6) + "-"
			+ Utils.doubleToString(cutPoints[i][j], 6) + "]");
	      }
	    }
	  }
	  attributes.addElement(new Attribute(getInputFormat().
					      attribute(i).name(),
					      attribValues));	  
	
      } else {
	attributes.addElement(getInputFormat().attribute(i).copy());
      }
    }    
    
    Instances outputFormat = 
      new Instances(getInputFormat().relationName(), attributes, 0);
    outputFormat.setClassIndex(classIndex);
    setOutputFormat(outputFormat);
  }

  /**
   * Convert a single instance over. The converted instance is added to 
   * the end of the output queue.
   *
   * @param instance the instance to convert
   */
  protected void convertInstance(Instance instance) {

    int index = 0;
    double [] vals = new double [outputFormatPeek().numAttributes()];
    // Copy and convert the values
    for(int i = 0; i < getInputFormat().numAttributes(); i++) {
      if (getInputFormat().attribute(i).isNumeric()) {
	int j;
	double currentVal = instance.value(i);
	if (cutPoints[i] == null) {
	  if (instance.isMissing(i)) {
	    vals[index] = Instance.missingValue();
	  } else {
	    vals[index] = 0;
	  }
	  index++;
	} else {
	    if (instance.isMissing(i)) {
	      vals[index] = Instance.missingValue();
	    } else {
	      for (j = 0; j < cutPoints[i].length; j++) {
		if (currentVal <= cutPoints[i][j]) {
		  break;
		}
	      }
              vals[index] = j;
	    }
	    index++;
	     
	}
      } else {
        vals[index] = instance.value(i);
	index++;
      }
    }
    
    Instance inst = null;
    if (instance instanceof SparseInstance) {
      inst = new SparseInstance(instance.weight(), vals);
    } else {
      inst = new Instance(instance.weight(), vals);
    }
    copyStringValues(inst, false, instance.dataset(), getInputStringIndex(),
                     getOutputFormat(), getOutputStringIndex());
    inst.setDataset(getOutputFormat());
    push(inst);
  }
  
  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }
  

  public String debugTipText(){
    return "Set debugging level between 0 and " + 
      (DynamicReductDiscretize.DEBUG_LEVELS-1);
  }
       
  /**
   * Get the value of reductIterations.
   * @return value of reductIterations.
   */
  public int getReductIterations() {
    return reductIterations;
  }
  
  /**
   * Set the value of reductIterations.
   * @param v  Value to assign to reductIterations.
   */
  public void setReductIterations(int  v) {
    this.reductIterations = v;
  }
  
  public String reductIterationsTipText(){
    return "number of iterations we should use for " 
      + "trying to obtain sets of proper reducts";
  }

  /**
   * Get the value of reductsRatio.
   * @return value of reductsRatio.
   */
  public double getReductsRatio() {
    return reductsRatio;
  }
  
  /**
   * @param v Value to assign to reductsRatio.  */
  public void setReductsRatio(double  v) {
    this.reductsRatio = v;
  }
  
  public String reductsRatioTipText() {
    return "fraction of the reducts for the subtables that should also be "
      + "reducts for the whole table";
  }

  
  /**
   * Get the value of minFraction.
   * @return value of minFraction.
   */
  public double getMinFraction() {
    return minFraction;
  }
  
  /**
   * Set the value of minFraction.
   * @param v  Value to assign to minFraction.
   */
  public void setMinFraction(double  v) {
    this.minFraction = v;
  }
  
  public String minFractionTipText(){
    return "the minimal size of the subtables, " 
      + "as a fraction of the original tables size, " 
      + "that should be used for stability calculation";
  }
  
  /**
   * Get the value of error.
   * @return value of error.
   */
  public double getError() {
    return error;
  }
  
  /**
   * Set the value of error.
   * @param v  Value to assign to error.
   */
  public void setError(double  v) {
    this.error = v;
  }
  
  public String errorTipText(){
    return "the error margin for the reductsRatio value";
  }

  /**
   * Get the value of maxIterations.
   * @return value of maxIterations.
   */
  public int getMaxIterations() {
    return maxIterations;
  }
  
  /**
   * Set the value of maxIterations.
   * @param v  Value to assign to maxIterations.
   */
  public void setMaxIterations(int  v) {
    this.maxIterations = v;
  }
  
  public String maxIterationsTipText(){
    return "the number of iterations in the binary search for a"
      + " proper subtable size we should perform";
  }
  
  /**
   * Get the value of numSubtables.
   * @return value of numSubtables.
   */
  public int getNumSubtables() {
    return numSubtables;
  }
  
  /**
   * Set the value of numSubtables.
   * @param v  Value to assign to numSubtables.
   */
  public void setNumSubtables(int  v) {
    this.numSubtables = v;
  }


  public String numSubtablesTipText(){
    return "Set the number of subtables that will be used for " 
      + "dynamic reduct construction";
  }
  
  /**
   * Get the value of fast.
   * @return value of fast.
   */
  public int getFast() {
    return fast;
  }
  
  /**
   * Set the value of fast.
   * @param v  Value to assign to fast.
   */
  public void setFast(int  v) {
    this.fast = v;
  }
  
  public String fastTipText(){
    return "Set the speed level (sacrifice stability)";
  }
  
  /**
   * Get the value of confidence.
   * @return value of confidence.
   */
  public double getConfidence() {
    return confidence;
  }
  
  /**
   * Set the value of confidence.
   * @param v  Value to assign to confidence.
   */
  public void setConfidence(double  v) {
    this.confidence = v;
  }

  public String confidenceTipText() {
    return "Set a value between 0 and 1 denoting the uncertainty allowed. " 
      + "(Low value => larger subtables)";
  }
  
  /**
   * Get the value of maxError.
   * @return value of maxError.
   */
  public double getMaxError() {
    return maxError;
  }
  
  /**
   * Set the value of maxError.
   * @param v  Value to assign to maxError.
   */
  public void setMaxError(double  v) {
    this.maxError = v;
  }
  
  public String maxErrorTipText() {
    return "Set the maximum acceptable hypothesis error value " 
      + "(default 0.05)";
  } 
  
  /**
     List the available options
   */

  public Enumeration listOptions() {

    Vector l = new Vector();

    l.addElement(new Option(debugTipText(),"D", 1, "-D <val>"));
    l.addElement(new Option(fastTipText(),"F",1,"-F <speed>"));
    l.addElement(new Option(reductsRatioTipText(),"T", 1,
			    "-T <reductsRatio>"));
    l.addElement(new Option(numSubtablesTipText(),"S", 1,
			    "-S <numSubtables>"));
    l.addElement(new Option(confidenceTipText(),"C", 1, 
			    "-C <confidence>"));
    l.addElement(new Option(reductIterationsTipText(),"R",1,
			    "-R <reductIterations>"));
    l.addElement(new Option(minFractionTipText(),"M",1,"-M <minFraction>"));

    return l.elements();
  }

  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */

  public void setOptions(String[] options) throws Exception{    

    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > DynamicReductDiscretize.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    + DynamicReductDiscretize.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()
    else {
      debug = 0;
    } // end of else

    String fastOption = Utils.getOption('F', options);
    if ( fastOption.length() > 0) {
      fast = Integer.parseInt(fastOption);
      if (fast  < 0 || fast > DynamicReductDiscretize.FAST) {
	throw new Exception("Illegal speed value. "
			    + "Choose a value between 0 and " 
			    + DynamicReductDiscretize.FAST);
      } // end of if ()
    } // end of if ()
    else {
      fast = DynamicReductDiscretize.FAST;
    } // end of else


    String reductsRatioOption = Utils.getOption('C', options);
    if ( reductsRatioOption.length() > 0) {
      reductsRatio = Double.parseDouble(reductsRatioOption);
      if (reductsRatio  < 0 || reductsRatio > 1) {
	throw new Exception("Illegal reductsRatio value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()

    String numSubtablesOption = Utils.getOption('S', options);
    if ( numSubtablesOption.length() > 0) {
      numSubtables = Integer.parseInt(numSubtablesOption);
      if (numSubtables  < 0) {
	throw new Exception("Illegal number of subtables. "
			    + "Choose a value larger than 0.");
      } // end of if ()
    } // end of if ()

    String reductIterationOption = Utils.getOption('R', options);
    if ( reductIterationOption.length() > 0) {
      reductIterations = Integer.parseInt(reductIterationOption);
    } // end of if ()

    String minFractionOption = Utils.getOption('M', options);
    if ( minFractionOption.length() > 0 ) {
      minFraction = Double.parseDouble(minFractionOption);
      if ( minFraction < 0 || minFraction > 1 ) {
	throw new Exception("Illegal minFraction value." + 
			    " Choose a value between 0 and 1");
      } // end of if ()
    } // end of if ()

    String confidenceOption = Utils.getOption('C', options);
    if ( confidenceOption.length() > 0) {
      confidence = Double.parseDouble(confidenceOption);
      if (confidence  < 0 || confidence > 1) {
	throw new Exception("Illegal confidence value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      confidence = DynamicReductDiscretize.CONFIDENCE;
    } // end of else

    String maxErrorOption = Utils.getOption('E', options);
    if ( maxErrorOption.length() > 0) {
      maxError = Double.parseDouble(maxErrorOption);
      if (maxError  < 0 || maxError > 1) {
	throw new Exception("Illegal maxError value. "
			    + "Choose a value between 0 and 1.");
      } // end of if ()
    } // end of if ()
    else {
      maxError = DynamicReductDiscretize.MAX_ERROR;
    } // end of else

  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] options = new String[10];
    int current = 0;

    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()

    if ( fast !=  DynamicReductDiscretize.FAST) {
      options[current++] = "-F";options[current++] = "" + fast;
    } // end of if ()

    if ( reductsRatio != DynamicReductDiscretize.REDUCTS_RATIO ) {
      options[current++] = "-T"; options[current++] = "" + reductsRatio;
    } // end of if ()

    if ( numSubtables != DynamicReductDiscretize.NUM_SUBTABLES ) {
      options[current++] = "-S"; options[current++] = "" + numSubtables;
    } // end of if ()

    if ( confidence != DynamicReductDiscretize.CONFIDENCE ) {
      options[current++] = "-C"; options[current++] = "" + confidence;
    } // end of if ()

    if ( maxError != DynamicReductDiscretize.MAX_ERROR ) {
      options[current++] = "-E"; options[current++] = "" + maxError;
    } // end of if ()

    if ( minFraction != .7 ) {
      options[current++] = "-M"; options[current++] = "" + minFraction;
    } // end of if ()
    
    if ( reductIterations != DynamicReductDiscretize.REDUCT_ITERATIONS ) {
      options[current++] = "-R"; options[current++] = "" + reductIterations;
    } // end of if ()
    
    
    /* Fill out the rest */

    while (current < options.length) {
      options[current++] = "";
    }

    return options;
  }

  public void printInfo() {
    System.out.println(this.toString());
  }

  /**
   * Returns a string describing this filter
   *
   * @return a description of the filter suitable for
   * displaying in the explorer/experimenter gui
   */

  public String globalInfo() {
    return "An instance filter that discretizes numeric"
      + " attributes in the dataset into nominal attributes."
      + " Discretization is done by using the Dynamic Reduct techniques" 
      + " put forth by Jan Bazan.\n"
      + " WARNING! You should definitely perform attribute selection" 
      + " prior to applying this filter!";
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain arguments to the filter: use -h for help
   */

  public static void main(String [] argv) {

    try {
      if (Utils.getFlag('b', argv)) {
	Filter.batchFilterFile(new DynamicReductDiscretize(), argv);
      } else {
	Filter.filterFile(new DynamicReductDiscretize(), argv);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }


}
