/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RSDiscretize
 * Copyleft (C) 2002 Ola Leifler
 *
 */

package weka.filters;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.roughset.*;

public class RSDiscretize extends Filter implements OptionHandler{
  
  static int DEBUG_LEVELS = 3;
  private int debug = 0;
  private boolean globalDiscretization = true;
  private boolean keepValues;

  /** Store the current cutpoints */
  protected double [][] cutPoints = null;
  
  private void localDiscretization(RoughInstances i) throws Exception{
      if ( ! RoughSet.sameDecision(i) ) {
	MDCutPoints cuts = new MDCutPoints(i);
	CutPoint maxCut = cuts.getMaxCut();
	bestCuts.addCut(maxCut);

	/* Create two new RoughInstances that contain the same header information */

	RoughInstances iLeft = new RoughInstances(i,0);
	RoughInstances iRight = new RoughInstances(i,0);
	for (Iterator iter = i.iterator();iter.hasNext(); ) {
	  RoughInstance inst = (RoughInstance)iter.next();
	  if ( inst.value(maxCut.getAttribute()) < maxCut.getCut() ) {
	    iLeft.add((RoughInstance)(inst.copy()));
	  }
	  else {
	    iRight.add((RoughInstance)(inst.copy()));
	  }
	} // end of for ()
	localDiscretization(iLeft);
	localDiscretization(iRight);
      
      }

  }

  
  private void globalDiscretization() throws Exception {
    Instances insts  = getInputFormat();
    MDCutPoints cuts = new MDCutPoints(insts);
    /**
       Should these classes be made modulo the nominal attributes in
       the dataset?  */

    SortedClasses c = new SortedClasses(insts,Attributes.nominalAttributes(insts));

    /**
       We do not allow much inconsistencies in the input data..
     */

    c.compact();
    CutPoint bestCut;
    while ( !c.sameDecisions() ) {
      bestCut = cuts.getMaxCut(c);
      if ( debug > 0 ) {
	System.out.println("The currently best cut is " 
			 + (bestCut == null ? "null" : bestCut.toString())); 
	if ( debug > 1) {
	  System.out.println(" from the set of cuts " + cuts.toString());
	} // end of if ()
	if ( debug > 2 ) {
	  System.out.println(" on the classes\n" + c.toString());
	} // end of if ()
	
      } // end of if ()
      bestCuts.addCut(bestCut);
      cuts.removeCut(bestCut);
      if ( debug > 0 ) {
	System.out.println((bestCut != null ? " divides\n" 
			   + c.toString() + "\n\ninto\n":""));
      } // end of if ()      
      c.splitClasses(bestCut);
      if ( debug > 0 && bestCut != null ) {
	c.printClasses();
      } // end of if ()
    } // end of while ()
    
  }
  
  protected MDCutPoints bestCuts;

  public RSDiscretize(){
    /** 
	Default values are set here:
	
     */ 
	bestCuts = null;
	keepValues = true;
  }
  
  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }
  
  public String debugTipText() { 
    return "Display debugging information. Choose a value between 0 and " 
      + RSDiscretize.DEBUG_LEVELS;
  }

    
  /**
   * Get the value of globalDiscretization.
   * @return value of globalDiscretization.
   */
  public boolean getGlobalDiscretization() {
    return globalDiscretization;
  }
    
  /**
   * Set the value of globalDiscretization.
   * @param v  Value to assign to globalDiscretization.
   */
  public void setGlobalDiscretization(boolean  v) {
    this.globalDiscretization = v;
  }
    
  public String globalDiscretizationTipText(){
    return "Compare the cuts with all the instances";
  }

  /**
   * Get the value of keepValues.
   * @return value of keepValues.
   */
  public boolean getKeepValues() {
    return keepValues;
  }
  
  /**
   * Set the value of keepValues.
   * @param v  Value to assign to keepValues.
   */
  public void setKeepValues(boolean  v) {
    this.keepValues = v;
  }
  
  public String keepValuesTipText(){
    return "Keep all non-discretized attribute values," 
      + " as compared to collapsing them";
  }


  protected void calculateCutPoints() throws Exception{
    bestCuts = new MDCutPoints();
    /* Use local discretization described on page 63 in RS Methods */
    if ( getGlobalDiscretization() ) {
      globalDiscretization();
    } // end of if ()
    else {
      RoughInstances r = new RoughInstances(getInputFormat());
      
      /**
	 We do not allow much inconsistencies in the input dataset
	 right now
      */

      r.compact();
      
      localDiscretization(r);      
    } // end of else	    
    if ( getKeepValues() ) {
      bestCuts.fillNonDiscretized(getInputFormat());
    } // end of if ()
    
    cutPoints = bestCuts.toDoubleArray(getInputFormat());
  }

  /**
   * Gets an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions() {

    Vector l = new Vector(7);

    l.addElement(new Option(debugTipText(),"D", 1, "-D <level>"));
    l.addElement(new Option(globalDiscretizationTipText(),"G", 0, "-G"));
    l.addElement(new Option(keepValuesTipText(),"K", 0, "-K"));

    return l.elements();
  }

  /**
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {


    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > RSDiscretize.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    + RSDiscretize.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()

    //    setGlobalDiscretization(Utils.getFlag('G', options));
    //    setKeepValues(Utils.getFlag('K', options));
    setKeepValues(true);
    if (getInputFormat() != null) {
      setInputFormat(getInputFormat());
    }
  }
  /**
   * Gets the current settings of the filter.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] options = new String [10];
    int current = 0;

    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()

    if (getGlobalDiscretization()) {
      options[current++] = "-G";
    }

//      if (getKeepValues()) {
//        options[current++] = "-K";
//      }

    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }



  /**
   * Input an instance for filtering. Ordinarily the instance is processed
   * and made available for output immediately. Some filters require all
   * instances be read before producing output.
   *
   * @param instance the input instance
   * @return true if the filtered instance may now be
   * collected with output().
   * @exception IllegalStateException if no input format has been defined.
   */
  public boolean input(Instance instance) {

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (m_NewBatch) {
      resetQueue();
      m_NewBatch = false;
    }
    
    if (cutPoints != null) {
      convertInstance(instance);
      return true;
    }
    else {
      bufferInput(instance);
      return false;
    } // end of else
    
  }


  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any instances contained in the object are ignored - only the
   * structure is required).
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the input format can't be set successfully
   */
  public boolean setInputFormat(Instances instanceInfo) throws Exception {
    super.setInputFormat(instanceInfo);
    bestCuts = null;  
    cutPoints = null;
    if (instanceInfo.classIndex() < 0) {
      throw new UnassignedClassException("Cannot use class-based " 
					 + "discretization: "
					 + "no class assigned to " 
					 + "the dataset");
    }
    if (!instanceInfo.classAttribute().isNominal()) {
      throw new UnsupportedClassTypeException("Supervised discretization " 
					      + "not possible:"
					      + " class is not nominal!");
    }
    
    // If we implement loading cutfiles, then load 
    //them here and set the output format
    return false;

  }

  /**
   * Signifies that this batch of input to the filter is finished. If the 
   * filter requires all instances prior to filtering, output() may now 
   * be called to retrieve the filtered instances.
   *
   * @return true if there are instances pending output
   * @exception IllegalStateException if no input structure has been defined
   */
  public boolean batchFinished() {

    try {
       
    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (cutPoints == null) {
      this.calculateCutPoints();
      if ( cutPoints != null && debug >= 1 ) {
	for (int i=0;i<cutPoints.length;i++) {
	  double [] tmp = cutPoints[i];
	  if ( tmp != null ) {	  
	    System.out.print("[");
	    for (int j=0;j<tmp.length;j++) {
	      System.out.print(tmp[j]+(j<tmp.length-1?",":""));
	    } // end of for ()
	    System.out.print("]");
	  } // end of if ()
	  System.out.println();
	} // end of for ()
	System.out.println("---------------------------------------------");
      } // end of if ()
      
      this.setOutputFormat();
      // If we implement saving cutfiles, save the cuts here

      // Convert pending input instances
      for(int i = 0; i < getInputFormat().numInstances(); i++) {
	convertInstance(getInputFormat().instance(i));
      }
    } 
    flushInput();

    m_NewBatch = true;
    
    } catch (Exception e) {
      e.printStackTrace();
    } // end of try-catch
    
    return (numPendingOutput() != 0);
  }


  /**
   * Set the output format. Takes the currently defined cutpoints and 
   * m_InputFormat and calls setOutputFormat(Instances) appropriately.
   */
  protected void setOutputFormat() {

    if (cutPoints == null) {
      setOutputFormat(null);
      return;
    }
    FastVector attributes = new FastVector(getInputFormat().numAttributes());
    int classIndex = getInputFormat().classIndex();
    for(int i = 0; i < getInputFormat().numAttributes(); i++) {
      if (getInputFormat().attribute(i).isNumeric()) {
	FastVector attribValues = new FastVector(1);
	if (cutPoints[i] == null) {
	  attribValues.addElement("All");
	} else {
	  for(int j = 0; j <= cutPoints[i].length; j++) {
	    if (j == 0) {
	      attribValues.addElement("(-inf-"
				      + Utils.doubleToString(cutPoints[i][j], 6) 
				      + "]");
	    } else if (j == cutPoints[i].length) {
	      attribValues.addElement("("
				      + Utils.doubleToString(cutPoints[i][j - 1], 6) 
				      + "-inf)");
	    } else {
	      attribValues.addElement("("
				      + Utils.doubleToString(cutPoints[i][j - 1], 6) 
				      + "-"
				      + Utils.doubleToString(cutPoints[i][j], 6) 
				      + "]");
	    }
	  }
	}
	attributes.addElement(new Attribute(getInputFormat().
					    attribute(i).name(),
					    attribValues));	  
	
      } else {
	attributes.addElement(getInputFormat().attribute(i).copy());
      }
    }    
    
    Instances outputFormat = 
      new Instances(getInputFormat().relationName(), attributes, 0);
    outputFormat.setClassIndex(classIndex);
    setOutputFormat(outputFormat);
  }

  /**
   * Convert a single instance over. The converted instance is added to 
   * the end of the output queue.
   *
   * @param instance the instance to convert
   */
  protected void convertInstance(Instance instance) {

    int index = 0;
    double [] vals = new double [outputFormatPeek().numAttributes()];
    // Copy and convert the values
    for(int i = 0; i < getInputFormat().numAttributes(); i++) {
      if (getInputFormat().attribute(i).isNumeric()) {
	int j;
	double currentVal = instance.value(i);
	if (cutPoints[i] == null) {
	  if (instance.isMissing(i)) {
	    vals[index] = Instance.missingValue();
	  } else {
	    vals[index] = 0;
	  }
	  index++;
	} else {
	    if (instance.isMissing(i)) {
	      vals[index] = Instance.missingValue();
	    } else {
	      for (j = 0; j < cutPoints[i].length; j++) {
		if (currentVal <= cutPoints[i][j]) {
		  break;
		}
	      }
              vals[index] = j;
	    }
	    index++;
	     
	}
      } else {
        vals[index] = instance.value(i);
	index++;
      }
    }
    
    Instance inst = null;
    if (instance instanceof SparseInstance) {
      inst = new SparseInstance(instance.weight(), vals);
    } else {
      inst = new Instance(instance.weight(), vals);
    }
    copyStringValues(inst, false, instance.dataset(), getInputStringIndex(),
                     getOutputFormat(), getOutputStringIndex());
    inst.setDataset(getOutputFormat());
    push(inst);
  }


  /**
   * Returns a string describing this filter
   *
   * @return a description of the filter suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {

    return "An instance filter that discretizes numeric"
      + " attributes in the dataset into nominal attributes."
      + " Discretization is done by using the general Rough Set techniques" 
      + " put forth by Nguyen Hung Son of the University of Poland";
  }


  /**
   * Main method for testing this class.
   *
   * @param argv should contain arguments to the filter: use -h for help
   */

  public static void main(String [] argv) {

    try {
      if (Utils.getFlag('b', argv)) {
	Filter.batchFilterFile(new RSDiscretize(), argv);
      } else {
	Filter.filterFile(new RSDiscretize(), argv);
      }
    } catch (Exception ex) {
      System.out.println(ex.getMessage());
    }
  }

}  


