/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    RSNumerizeNominal.java
 *    Copyright (C) 1999 Len Trigg
 *
 */


package weka.filters;

import java.io.*;
import java.util.*;
import weka.core.*;

/** 
 * Creates numerical attributes of nominal ones, in case we wish to
 * discretize them also
 *
 * @author Ola Leifler
 * @version $Revision: 1.1 $ */
public class RSNumerizeNominal extends Filter implements OptionHandler {

  private Range range = new Range();
  private int debug;
  private static int DEBUG_LEVELS = 3;

  public String [] getOptions() {
    
    String [] options = new String [10];
    int current = 0;
    
    if ( debug > 0 ) {
      options[current++] = "-D"; options[current++] = "" + debug;
    } // end of if ()
    
    if ( !range.getRanges().equals("first-last") ) {
      options[current++] = "-R";options[current++] = range.getRanges();
    } // end of if ()
    
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }
  
  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }

  public String debugTipText(){
    return "Set debugging level between 0 and " + RSNumerizeNominal.DEBUG_LEVELS;
  }
  

  /**
     The get- set- and TipText-methods must exist in order for the
     algorithm to have it's fields modified in the WEKA GUI. Also, the
     arguments and return types of the get- and set-methods must be of
     a type that already has an associated "registered editor" (see
     Interface java.beans.PropertyEditor). 
   */

  /**
   * Returns the tip text for this property
   *
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String attributeIndicesTipText() {
    return "Specify range of attributes to act on."
      + " This is a comma separated list of attribute indices, with"
      + " \"first\" and \"last\" valid values. Specify an inclusive"
      + " range with \"-\". E.g: \"first-3,5,6-10,last\".";
  }

  /**
   * Gets the current range selection
   *
   * @return a string containing a comma separated list of ranges
   */
  public String getAttributeIndices() {

    return range.getRanges();
  }

  /**
   * Sets which attributes are to be Discretized (only numeric
   * attributes among the selection will be Discretized).
   *
   * @param rangeList a string representing the list of attributes. Since
   * the string will typically come from a user, attributes are indexed from
   * 1. <br>
   * eg: first-3,5,6-last
   * @exception IllegalArgumentException if an invalid range list is supplied 
   */
  public void setAttributeIndices(String rangeList) {

    range.setRanges(rangeList);
  }
  
  /**
   * Gets an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions() {

    Vector l = new Vector(7);

    l.addElement(new Option(debugTipText(),"D", 1, "-D <level>"));
    l.addElement(new Option(attributeIndicesTipText(),"R",1,"-R <range>"));

    return l.elements();
  }

  public void setOptions(String[] options) throws Exception {
    String debugOption = Utils.getOption('D', options);
    if ( debugOption.length() > 0) {
      debug = Integer.parseInt(debugOption);
      if (debug  < 0 || debug > RSNumerizeNominal.DEBUG_LEVELS ) {
	throw new Exception("Illegal debugging level. "
			    + "Choose a value between 0 and " 
			    + RSNumerizeNominal.DEBUG_LEVELS);
      } // end of if ()
    } // end of if ()
    
    String convertList = Utils.getOption('R', options);
    if (convertList.length() != 0) {
      range.setRanges(convertList);
    } else {
      range.setRanges("first-last");
    }
    
  }

  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any instances contained in the object are ignored - only the
   * structure is required).
   * @return true if the outputFormat may be collected immediately
   */
  public boolean setInputFormat(Instances instanceInfo) throws Exception {

    super.setInputFormat(instanceInfo);
    range.setUpper(instanceInfo.numAttributes()-1);
    setOutputFormat();
    return true;
  }

  protected void setOutputFormat() throws Exception {
    Instances oldInsts = getInputFormat();
    FastVector attributes = new FastVector(oldInsts.numAttributes());
    int classIndex = oldInsts.classIndex();
    for (int i=0;i<oldInsts.numAttributes();i++) {
      if ( range.isInRange(i) ) {
	attributes.addElement(new Attribute(oldInsts.attribute(i).name()));
      } else {
	attributes.addElement(oldInsts.attribute(i).copy());
      } // end of else
    } // end of for ()
    Instances newInsts = new Instances(oldInsts.relationName(),attributes,0);
    if ( debug > 0 ) {
      System.out.println("Output format set to\n" + newInsts.toString());      
    } // end of if ()
  
    newInsts.setClassIndex(classIndex);
    setOutputFormat(newInsts);
    
    } 


  /**
   * Input an instance for filtering. Ordinarily the instance is processed
   * and made available for output immediately. Some filters require all
   * instances be read before producing output.
   *
   * @param instance the input instance
   * @return true if the filtered instance may now be
   * collected with output().
   * @exception IllegalStateException if no input format has been defined.
   */
  public boolean input(Instance instance) {

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (m_NewBatch) {
      resetQueue();
      m_NewBatch = false;
    }
    push((Instance)instance.copy());
    return true;
  }

  /**
   * Returns a string describing this filter
   *
   * @return a description of the filter suitable for
   * displaying in the explorer/experimenter gui
   */

  public String globalInfo() {
    return 
      "Create numerical attributes of nominal ones, " 
      + "to be used prior to discretization";
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain arguments to the filter: use -h for help
   */
  public static void main(String [] argv) {
    
    try {
      if (Utils.getFlag('b', argv)) {
	Filter.batchFilterFile(new RSNumerizeNominal(), argv);
      } else {
	Filter.filterFile(new RSNumerizeNominal(), argv);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}

