/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    RSReplaceMissingValues.java
 *    Copyright (C) 1999 Eibe Frank
 *
 */


package weka.filters;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.roughset.*;

/** 
 * Replaces all missing values for nominal and numeric attributes in a 
 * dataset with the modes and means from the training data.
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz) 
 * @version $Revision: 1.1 $
 */

public class RSReplaceMissingValues extends Filter {

  /** The mean values of numeric attributes are collected in a vector
      that is used for conversion.
  */

  private double[] conversionVector = null;

  /**
     This hashtable has a number of sub-hashtables for each decision
     value that keeps track of how many instances have a certain value
     for a certain attribute, given the decision value.
   */

  private Hashtable counts = null;

  /**
     Debugging level, used for development purposes alone.
   */

  private int debug = 0; 

  public RSReplaceMissingValues(){
    debug = 0;
  }

  public RSReplaceMissingValues(int d){
    debug = d;
  }

  /**
   * Get the value of debug.
   * @return value of debug.
   */
  public int getDebug() {
    return debug;
  }
  
  /**
   * Set the value of debug.
   * @param v  Value to assign to debug.
   */
  public void setDebug(int  v) {
    this.debug = v;
  }
  
  public String debugTipText(){
    return "Set debugging level";
  }


  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input 
   * instance structure (any instances contained in the object are 
   * ignored - only the structure is required).
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the input format can't be set 
   * successfully
   */
  public boolean setInputFormat(Instances instanceInfo) 
       throws Exception {

    super.setInputFormat(instanceInfo);
    setOutputFormat(instanceInfo);
    conversionVector = null;
    return true;
  }

  /**
   * Input an instance for filtering. Filter requires all
   * training instances be read before producing output.
   *
   * @param instance the input instance
   * @return true if the filtered instance may now be
   * collected with output().
   * @exception IllegalStateException if no input format has been set.
   */

  public boolean input(Instance instance) {

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (m_NewBatch) {
      resetQueue();
      m_NewBatch = false;
    }
    if ( conversionVector != null ) {
      convertInstance(instance);
      return true;
    } // end of if ()
    else {
      bufferInput(instance);
      return false;
    } // end of else    
  }

  /**
   * Signify that this batch of input to the filter is finished. 
   * If the filter requires all instances prior to filtering,
   * output() may now be called to retrieve the filtered instances.
   *
   * @return true if there are instances pending output
   * @exception IllegalStateException if no input structure has been defined
   */
  public boolean batchFinished() throws Exception {

    if (getInputFormat() == null) {
      throw new IllegalStateException("No input instance format defined");
    }
    if (getInputFormat().classIndex() < 0) {
      throw new UnassignedClassException("Cannot use class-based " 
					 + "discretization: "
					 + "no class assigned to " 
					 + "the dataset");
    }
    if (!getInputFormat().classAttribute().isNominal()) {
      throw new UnsupportedClassTypeException("Supervised discretization " 
					      + "not possible:"
					      + " class is not nominal!");
    }

    try {
       
    if (conversionVector == null) {
      // Compute conversion matrix
      double sumOfWeights =  getInputFormat().sumOfWeights();

      counts = new Hashtable(getInputFormat().numClasses() * 
				       getInputFormat().numAttributes());
      
      double[] sums = new double[getInputFormat().numAttributes()];
      Arrays.fill(sums,sumOfWeights);
      Integer zero = new Integer(0);

      double[] results = new double[getInputFormat().numAttributes()];
      Attribute cAttr = getInputFormat().classAttribute();
      for (int j = 0; j < getInputFormat().numInstances(); j++) {
	Instance inst = getInputFormat().instance(j);
	for (int i = 0; i < inst.numValues(); i++) {
	  if (!inst.isMissingSparse(i)) {
	    double value = inst.valueSparse(i);
	    Attribute attr = inst.attributeSparse(i);
	    if (attr.isNominal() 
		&& !inst.classIsMissing() 
		&& !attr.equals(inst.classAttribute()) ) {
	      /**
		 This is where we add information about what attribute
		 values are most frequent int each decision class
	      */
	      
	      AttributeValuePair p = new AttributeValuePair(inst);
	      Hashtable h = null;
	      ArrayList l = null;
	      if ( counts.containsKey(p) ) {
		h = (Hashtable)counts.get(p);
	      } // end of if ()
	      else {
		h = new Hashtable(getInputFormat().numAttributes());
	      } // end of else
	      if ( h.containsKey(attr) ) {
		l = (ArrayList)h.get(attr);
	      } // end of if ()
	      else {
		l = new ArrayList(attr.numValues());
	      } // end of else
	      for (int n=l.size();n<=(int)value;n++) {
		l.add(n,zero);
	      } // end of for ()
	      l.set((int)value,
		    new Integer(((Integer)l.get((int)value)).intValue()+1));  
	      if ( debug >= 2 ) {
		System.out.println(inst.toString() + " => " + attr.name() 
				   + " = " + (int)value + ", Frequency: " 
				   + ((Integer)l.get((int)value)).toString());
	      } // end of if ()
	      h.put(attr,l);
	      counts.put(p,h);
	    } else if (inst.attributeSparse(i).isNumeric()) {
	      results[i] += inst.weight() * value;
	    }
	  } else {
	    if (inst.attributeSparse(i).isNumeric()) {
	      sums[i] -= inst.weight();
	    }
	  }
	}
      }
      
      Hashtable newTable = new Hashtable(counts);
      for (Iterator i = counts.keySet().iterator();i.hasNext();) {
	AttributeValuePair p = (AttributeValuePair)i.next();
	Hashtable h = (Hashtable)counts.get(p);
	Hashtable copy = new Hashtable(h);
	if ( debug >= 2 ) {
	  System.out.println("Investigating class " + p);
	} // end of if ()	
	for (Iterator j = h.keySet().iterator();j.hasNext();) {
	  Attribute attr = (Attribute)j.next();
	  if ( !attr.equals(getInputFormat().classAttribute())) {
	    ArrayList a = new ArrayList((ArrayList)h.get(attr));
	    Integer bestValue = new Integer(0);
	    int maxOccurences = 0;
	    if ( debug >= 2 ) {
	      System.out.println("Attribute " + attr.name());
	    } // end of if ()	    
	    for (int k = 0;k < a.size();k++) {
	      int val = ((Integer)a.get(k)).intValue();
	      if ( debug >= 2 ) {
		System.out.println("Value " + k + ", frequency " + val);
	      } // end of if ()
	      if ( val  > maxOccurences ) {
		bestValue = new Integer(k);
		maxOccurences = val;
	      } // end of if ()
	    } // end of for ()
	    if ( debug >= 2 ) {
	      System.out.println("Adding hashtable entry: " 
				 + attr.name() + " => " 
				 + bestValue.toString());
	    } // end of if ()
	    copy.put(attr,bestValue);
	  } else {
	    copy.remove(attr);
	  } // end of else
	  

	} // end of for ()
	newTable.put(p,copy);
      } // end for()
      counts = newTable;

      if ( debug >= 1 ) {
	for (Iterator i = counts.keySet().iterator() ; i.hasNext() ; ) {
	  AttributeValuePair p = (AttributeValuePair)i.next();
	  System.out.println("Hashtable entry:\ndecision = " 
			     + p.toString() + ", best attribute values = " );
	  Hashtable h = (Hashtable)counts.get(p);
	  for (Iterator j = h.keySet().iterator() ; 
	       j.hasNext() ; ) {
	    Attribute a = (Attribute)j.next();
	    if ( true ) {
	      System.out.println(a.name() + " = " 
				 + ((Integer)(h.get(a))).
				 toString());
	    } // end of if ()
	  } // end of for ()
	} // end of for ()
      } // end of if ()
      

      conversionVector = new double[getInputFormat().numAttributes()];
      Arrays.fill(conversionVector,0);
      for (int i = 0; i < getInputFormat().numAttributes(); i++) {
	if (getInputFormat().attribute(i).isNumeric()) {
	  if (Utils.gr(sums[i], 0)) {
	    conversionVector[i] = results[i] / sums[i];
	  }
	}
      }
      

      // Convert pending input instances
      for(int i = 0; i < getInputFormat().numInstances(); i++) {
	convertInstance(getInputFormat().instance(i));
      }
    } 

    } catch (Exception e) {
      e.printStackTrace();
    } // end of try-catch
    

    // Free memory
    flushInput();

    m_NewBatch = true;
    return (numPendingOutput() != 0);
  }

  /**
   * Convert a single instance over. The converted instance is 
   * added to the end of the output queue.
   *
   * @param instance the instance to convert
   */
  private void convertInstance(Instance instance) {
  
    Instance inst = null;
    /**
       We will skip entirely instances where the class values are missing
    */
    
    if ( instance.classIsMissing() ) {
      return;
    } // end of if ()
    
    if (instance instanceof SparseInstance) {
      double []vals = new double[instance.numValues()];
      int []indices = new int[instance.numValues()];
      int num = 0;
      for (int j = 0; j < instance.numValues(); j++) {
	
	if (instance.isMissingSparse(j) &&
	     instance.attributeSparse(j).isNumeric()) {
	  if (conversionVector[instance.index(j)] != 0.0) {
	    vals[num] = conversionVector[instance.index(j)];
	    indices[num] = instance.index(j);
	    num++;
	  } 
	} else if ( instance.isMissingSparse(j) && 
		    instance.attributeSparse(j).isNominal() ) {
	  if (conversionVector[instance.index(j)] != 0.0) {
	    Hashtable h = (Hashtable)counts.get(new AttributeValuePair(instance));
	    Integer val = (Integer)h.get(instance.attributeSparse(num)); 
	    vals[num] = val.doubleValue();
	    indices[num] = instance.index(j);
	    num++;
	  } 
	  
	} else {
	  vals[num] = instance.valueSparse(j);
	  indices[num] = instance.index(j);
	  num++;
	}
      } 
      if (num == instance.numValues()) {
	inst = new SparseInstance(instance.weight(), vals, indices,
                                  instance.numAttributes());
      } else {
	double []tempVals = new double[num];
	int []tempInd = new int[num];
	System.arraycopy(vals, 0, tempVals, 0, num);
	System.arraycopy(indices, 0, tempInd, 0, num);
	inst = new SparseInstance(instance.weight(), tempVals, tempInd,
                                  instance.numAttributes());
      }
    } else {
      double []vals = new double[getInputFormat().numAttributes()];
      for (int j = 0; j < instance.numAttributes(); j++) {
	if (instance.isMissing(j) &&
	    getInputFormat().attribute(j).isNumeric()) {
	  vals[j] = conversionVector[j]; 
	} else if ( instance.isMissing(j) &&
		    getInputFormat().attribute(j).isNominal() ) {	  
	  Hashtable h = (Hashtable)counts.get(new AttributeValuePair(instance));
	  Integer val = (Integer)h.get(instance.attribute(j)); 
	  if ( val == null ) {
	    System.out.println("No values present for " 
			       + instance.attribute(j).name() 
			       + " for decision " 
			       + new AttributeValuePair(instance).toString());
	  } // end of if ()
	  
	  vals[j] = val.doubleValue();
	} else {
	  vals[j] = instance.value(j);
	}
      } 
      inst = new Instance(instance.weight(), vals);
    } 
    inst.setDataset(instance.dataset());
    push(inst);
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain arguments to the filter: 
   * use -h for help
   */
  public static void main(String [] argv) {

    try {
      int debug = 2;
      if (Utils.getFlag('b', argv)) {
 	Filter.batchFilterFile(new RSReplaceMissingValues(debug), argv);
      } else {
	Filter.filterFile(new RSReplaceMissingValues(debug), argv);
      }
    } catch (Exception ex) {
      System.out.println(ex.getMessage());
    }
  }


}
