/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Rule.java
 *    Copyleft (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.io.*;
import java.util.*;
import weka.core.*;

/**
 * Represents a decision rule
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.1
 */
public class Rule extends Object {

  private AttributeValuePairs conditions;
  private RuleSet ruleSet;  
  private Collection matching;
  private Collection supporting;
  private double stability;

  /* Stupid name, but it's supposed to sort out all those attributes that */

  private double numNegative(AttributeValuePair p) {
    double sum = 0;
    Rule r = new Rule(this);
    r.removeCondition(p);
    Enumeration e = this.getInstances().enumerateInstances();
    while ( e.hasMoreElements() ) {
      Instance i = (Instance)e.nextElement();
      sum += ((r.matches(i) && !r.supports(i)) ? 1 : 0);
    } // end of while ()
    return sum;
  }

  public Rule(RuleSet r, AttributeValuePairs c) {
    ruleSet = r;
    conditions = c;
    matching = null;
    supporting = null;
    stability = -1;
  }

  public Rule(Rule r) {
    ruleSet = r.getRuleSet();
    conditions = new AttributeValuePairs(r.getConditions());
    matching = new ArrayList(r.getMatchingInstances());
    supporting = new ArrayList(r.getSupportingInstances());
    stability = -1;
  }

  public Rule(Instance i, Reduct r) {
    ruleSet = null;
    conditions = new AttributeValuePairs(i,r.getAttributes());
    matching = null;
    supporting = null;
    stability = -1;
  }

  /**
   * Get the value of ruleSet.
   * @return value of ruleSet.
   */

  public RuleSet getRuleSet() {
    return ruleSet;
  }
  
  /**
   * Set the value of ruleSet.
   * @param v  Value to assign to ruleSet.
   */
  public void setRuleSet(RuleSet  v) {
    this.ruleSet = v;
    matching = null;
    supporting = null;
  }
  
  public AttributeValuePairs getConditions() {
    return conditions;
  }
  
  public void removeCondition(AttributeValuePair p) {
    conditions.removePair(p);
    matching = null;
    supporting = null;
  }

  public void removeCondition(int i) {
    conditions.removePair(i);
    matching = null;
    supporting = null;
  }

  public void addCondition(AttributeValuePair p) {
    conditions.addPair(p);
    matching = null;
    supporting = null;
  }

  public double getDecisionValue() {
    return getDecision().getValue();
  }
  
  public AttributeValuePair getDecision() {
    return getRuleSet().getDecision();
  }

  /* this measure was introduced by the LERS algorithm  */

  public int getSupport() {
    return this.numSupporting() * conditions.size();
  }

  public Instances getInstances() {
    return getRuleSet().getInstances();
  }
  
  public Collection getMatchingInstances() {
    if ( matching == null  ) {    
      Instances insts = this.getInstances();
      Enumeration e = insts.enumerateInstances();
      ArrayList a = new ArrayList();
      while ( e.hasMoreElements() ) {
	Instance i = (Instance)e.nextElement();
	if ( this.matches(i) ) {
	  a.add(i);
	} // end of if ()
      } // end of while ()
      matching = a;
    } // end of if ()
    return matching;
  }
  
  public Collection getSupportingInstances() {
    if ( supporting == null ) {    
      Instances insts = this.getInstances();
      Enumeration e = insts.enumerateInstances();
      ArrayList a = new ArrayList();
      while ( e.hasMoreElements() ) {
	Instance i = (Instance)e.nextElement();
	if ( this.supports(i) ) {
	  a.add(i);
	} // end of if ()
      } // end of while ()
      supporting = a;
    } // end of if ()
    return supporting;
  }

  public int numMatching() {
    return this.getMatchingInstances().size();
  }

  public boolean matches(Instance i) {
    return conditions.matches(i);
  }

  public boolean supports(Instance i) {
    return this.matches(i) && ((i.classValue() == getDecisionValue()) || 
			       i.classIsMissing());
  }

  public int numSupporting() {
    return this.getSupportingInstances().size();
  }

  public double partialSupport(Instance i) {
    return (double)(conditions.getPartialMatch(i)*
		    this.numSupporting()*conditions.size());
  }

  public double conditionalProbability() {
    return (double)(this.numSupporting()/this.numMatching());
  }

  public double stabilityStrength() {
    if ( stability == -1 ) {
      double numerator = -1;
      double sum =  this.numMatching() - this.numSupporting();
      for (int j=0;j<conditions.size();j++) {
	AttributeValuePair p = conditions.get(j);
	double numNegative = numNegative(p);
	numerator += Math.pow(2,numNegative);
	sum += numNegative;
      } // end of for ()
      double denominator = Math.pow(2,sum);
      stability = numerator/denominator;
    } // end of if ()
    return stability;
  }

  public void approximateRule(double consistency) {
    double maxCoeff = this.conditionalProbability(),coeff = 0;
    while ( maxCoeff >= consistency ) {
      int i = 0,maxIndex = -1;
      Iterator iter = this.getConditions().iterator();
      maxCoeff = 0;
      while ( iter.hasNext() ) {
	AttributeValuePair p = (AttributeValuePair)iter.next();
	Rule approximation = new Rule(this);
	approximation.removeCondition(p);
	coeff = approximation.conditionalProbability();
	if ( coeff > maxCoeff ) {
	  maxCoeff = coeff;
	  maxIndex = i;
	} // end of if ()
	i++;
      } // end of while ()
      if ( maxCoeff >= consistency && maxIndex >= 0 ) {
	this.removeCondition(maxIndex);
      } // end of if ()
    } // end of while ()
  }

  public String toString() {
    StringBuffer s = new StringBuffer();
    s.append(conditions.ruleString());
    s.append(" (" + this.numSupporting() + "/" + this.numMatching() + ")");
    return s.toString();
  }

  public boolean equals(Object o) {
    return ((Rule)o).toString().equals(this.toString());
  }

  public int hashCode() {
    return this.toString().hashCode();
  }


  


} 

