/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    RuleSet.java
 *    Copyright (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import weka.core.*;

/**
 * A set of rules that all as a consequent have a common decision value. 
 *
 * @author <a href="mailto:olale@aste1.ida.liu.se">Ola Leifler</a>
 * @version 1.1
 */

class RuleSet extends Object implements Comparable {

  private AttributeValuePair decision;
  private Set rules;
  private Rules parentSet;
  
  public RuleSet(){
    rules = new HashSet();
    decision = null;
    parentSet = null;
  }

  public RuleSet(Rules r, AttributeValuePair d){
    rules = new HashSet();
    decision = d;
    parentSet = r;
  }

  public RuleSet(Rules r, LocalCovering c, AttributeValuePair d){
    rules = new HashSet(c.size());
    decision = d;
    parentSet = r;
    addRules(c);
  }

  public RuleSet(RuleSet r){
    rules = new HashSet(r.getRules());
    parentSet = r.getParentSet();
    decision = r.getDecision();
  }

  /**
   * Get the value of parentSet.
   * @return value of parentSet.
   */
  public Rules getParentSet() {
    return parentSet;
  }
  
  /**
   * Set the value of parentSet.
   * @param v  Value to assign to parentSet.
   */
  public void setParentSet(Rules  v) {
    this.parentSet = v;
  }
    
  /**
   * Get the value of decision.
   * @return value of decision.
   */
  public AttributeValuePair getDecision() {
    return decision;
  }
  
  /**
   * Set the value of decision.
   * @param v  Value to assign to decision.
   */
  public void setDecision(AttributeValuePair  v) {
    this.decision = v;
  }
  
  public Collection getRules(){
    return rules;
  }

  public void addRule(Rule r){
    r.setRuleSet(this);
    rules.add(r);
  }

  public void addRules(LocalCovering c) {
    Iterator iter = c.iterator();
    while ( iter.hasNext() ) {
      AttributeValuePairs p = (AttributeValuePairs)iter.next();
      this.addRule(new Rule(this,p));      
    } // end of while ()
  }

  public void addAll(RuleSet s) {
    Iterator i = s.iterator();
    while ( i.hasNext() ) {
      Rule r = (Rule)i.next();
      rules.add(r);
    } // end of while ()
  }

  public void removeRule(Rule r) {
    rules.remove(r);
  }

  public AttributeValuePair getDecisionPair() {
    return decision;
  }

  public double getDecisionValue(){
    return decision.getValue();
  }

  public double size() {
    return (double)rules.size();
  }

  /**
     A very shallow comparison but it'll have to do for now.
   */

  public boolean equals(Object o) {
    return decision.equals(((RuleSet)o).getDecisionPair());
  }

  public int compareTo(Object o) {
    return decision.compareTo(((RuleSet)o).getDecisionPair());
  }
  
  public Iterator iterator() {
    return rules.iterator();
  }

  public Instances getInstances() {
    return parentSet.getInstances();
  }

  public double numMatching() {
    double sum = 0;
    for (int i=0;i<this.getInstances().numInstances();i++) {
      Instance inst = getInstances().instance(i);
      Iterator iter = rules.iterator();
      while ( iter.hasNext() ) {
	Rule nextRule = (Rule)iter.next();
	if ( nextRule.matches(inst) ) {
	  sum++;
	  break;
	} // end of if ()
      } // end of while ()
    } // end of for
    return sum;
  }

  public double numMatching(Instance i) {
    return (double)this.getMatchingRules(i).size();
  }

  private List getMatchingRules(Instance i) {
    Iterator iter = rules.iterator();
    ArrayList res = new ArrayList();
    while ( iter.hasNext() ) {
      Rule nextRule = (Rule)iter.next();
      if ( nextRule.matches(i) ) {
	res.add(nextRule);
      } // end of if ()
    } // end of while ()
    return res;
  }

  public double numSupporting() {
    double sum = 0;
    for (int i=0;i<this.getInstances().numInstances();i++) {
      Instance inst = getInstances().instance(i);
      Iterator iter = rules.iterator();
      while ( iter.hasNext() ) {
	Rule nextRule = (Rule)iter.next();
	if ( nextRule.supports(inst) ) {
	  sum++;
	  break;
	} // end of if ()
      } // end of while ()
    } // end of for
    return sum;
  }

  /* How many instances have the same decision as the rules in this class? */

  public double numWithDecision() {
    Enumeration e = parentSet.getInstances().enumerateInstances();
    double sum = 0; 
    while ( e.hasMoreElements() ) {
      Instance i = (Instance)e.nextElement();
      sum += ((i.classAttribute().equals(decision.getAttribute())) && 
	      (i.classValue() == decision.getValue()) ? 1 : 0);
    } // end of while ()
    return sum;
  }
  /* My personal interpretation of rule negotiation is this.. */

  public double simpleStrength(Instance i){
    return this.numMatching(i)/this.size();
  }

  public double maximalStrength(Instance i){
    double maxSupporting = 0;
    Iterator iter = this.getMatchingRules(i).iterator();
    while ( iter.hasNext() ) {
      maxSupporting = Math.max(maxSupporting,
			       ((Rule)iter.next()).numSupporting());
    } // end of while ()
    return maxSupporting/this.numWithDecision();
  }


  public double globalStrength(Instance i) {
    double supporting = 0;
    Iterator iter = this.getMatchingRules(i).iterator();
    while ( iter.hasNext() ) {
      supporting += ((Rule)iter.next()).numSupporting();
    } // end of while ()
    return supporting/this.numWithDecision();
  }

  public double stabilityStrength(Instance i) {
    List l = this.getMatchingRules(i);
    int numRules = l.size();
    double maxStability = 0;
    for (int j=0;j<numRules;j++) {
      maxStability = Math.max(maxStability,((Rule)l.get(j)).stabilityStrength());
    } // end of for ()
    return maxStability;
  }

  public double globalStabilityStrength(Instance i) {
    List l = this.getMatchingRules(i);
    int numRules = l.size();
    double sum = 0;
    for (int j=0;j<numRules;j++) {
      double stabilityStrength = ((Rule)l.get(j)).stabilityStrength();
      sum = (stabilityStrength + sum) - (stabilityStrength * sum); 
    } // end of for ()
    return sum;
  }

  public double conditionalStrength(Instance i){
    List l = this.getMatchingRules(i);
    int numRules = l.size();
    double maxSupport = 0;
    for (int j=0;j<numRules;j++) {
      maxSupport = Math.max(maxSupport,((Rule)l.get(j)).conditionalProbability());
    } // end of for ()
    return maxSupport;
  }

  public double partialSupportStrength(Instance i) {
    List r = this.getMatchingRules(i);
    Iterator iter = r.iterator();
    double sum = 0;
    while ( iter.hasNext() ) {
      Rule rule = (Rule)iter.next();
      sum += rule.partialSupport(i);
    } // end of while ()

    /* How large part of this RuleSet matches a certain instance? */

    return sum;
  }

  public double conditionalProbability(){
    return this.numMatching()/this.numSupporting();
  }

  public void approximateRules(double consistency) {
    Iterator iter = rules.iterator();
    HashSet h = new HashSet(rules.size());
    while ( iter.hasNext() ) {
      Rule r = (Rule)iter.next();
      iter.remove();
      r.approximateRule(consistency);
      h.add(r);
    } // end of for ()
    rules.addAll(h);
  }

  public void reduceRuleSet(double consistency) {
    double maxCoeff = this.conditionalProbability(),coeff = 0;
    while ( maxCoeff >= consistency ) {
      int i = 0;
      Rule maxRule = null;
      Iterator iter = iterator();
      maxCoeff = 0;
      while ( iter.hasNext() ) {
	Rule p = (Rule)iter.next();
	RuleSet approximation = new RuleSet(this);
	approximation.removeRule(p);
	coeff = approximation.numMatching()/approximation.numWithDecision();
	if ( coeff > maxCoeff ) {
	  maxCoeff = coeff;
	  maxRule = p;
	} // end of if ()
	i++;
      } // end of while ()
      if ( maxCoeff >= consistency && maxRule != null ) {
	this.removeRule(maxRule);
      } // end of if ()
    } // end of while ()
  }

  public String toString() {
    Iterator i = rules.iterator();
    StringBuffer s = new StringBuffer();
    s.append("=====================" 
	     + decision.toString() 
	     + "=====================\n");
    while ( i.hasNext() ) {
      s.append(((Rule)i.next()).toString());
      s.append("\n");
    } // end of while ()
    s.append("----------------------(" 
	     + numSupporting()
	     + "/" 
	     + numMatching() 
	     + ")----------------------\n");
    return s.toString();
  }


}
