/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * SubtableChooser
 * Copyleft (C) 2002 Ola Leifler
 *
 */

package weka.roughset;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

import drasys.or.prob.NormalDistribution;

import weka.core.Instances;
import weka.core.Statistics;
import weka.core.Utils;

public class SubtableChooser {

  private double maxVal;
  private int numChoices;
  private double[] probabilities;
  private int minSize;
  private int tableSize;
  private Instances instances;
  private HashSet h;

  private boolean containsEnough(int size){
    int numWithSize = 0;
    for (Iterator iter=h.iterator();iter.hasNext();) {
      IntegerSet s = (IntegerSet)iter.next();
      numWithSize += (s.size()==size?1:0);
    } // end of for ()
    return numWithSize == Utils.nChooseK(tableSize,size).intValue();

  }

  static int numChoices(double confidence, double maxError){
    return  (int)Math.round(Math.pow((new NormalDistribution().inverseCdf(1-(confidence/2))/
				      (2*maxError)),2));
  }

  /* Function in chapter 8 of Bazan98 - Dynamic Comparison */

  public SubtableChooser(Instances n, int minTableSize, double c, double m) 
    throws Exception {
    instances = n;
    numChoices = numChoices(c,m);    
    minSize = minTableSize;
    tableSize = instances.numInstances();
    if ( minSize > tableSize ) {
      throw new Exception("Too small subtable size!");
    } // end of if ()

    int gap = tableSize-minSize;
    BigDecimal numCombinations = new BigDecimal(BigInteger.ZERO);
    BigDecimal next;
    ArrayList combinations;

    if (gap > 0) {    
      combinations = new ArrayList(gap);
      for (int k=0;k<=gap;k++) {
	next = new BigDecimal(Utils.nChooseK(tableSize,k+minSize));
	combinations.add(k,next);
	numCombinations = numCombinations.add(next);
      } // end of for ()
      if ( new BigDecimal(numChoices).compareTo(numCombinations) > 0 ) {
	throw new Exception("Too small subtable size!");
      } // end of if ()
      probabilities = new double[gap+1];
      maxVal = 0;
      for (int i=0;i<=gap;i++) {
	probabilities[i] = 
	  ((BigDecimal)combinations.get(i)).divide(numCombinations,53,0).
	  doubleValue();
	maxVal += probabilities[i];
      } // end of for ()
    } // end of if ()
    else {
      probabilities = new double[tableSize];
    } // end of else
    h = new HashSet(numChoices);
  }

  /** Choose a static mount of subtables */

  public SubtableChooser(Instances i, int size, int numTables) 
    throws Exception {
    this(i,size,0,0);
    numChoices = numTables;
  }
  
  public boolean hasNext() {
    return numChoices > 0;
  }
  
  public Instances next() {
    numChoices--;
    int size = minSize;
    do {
      double val = maxVal * Math.random();
      size = minSize;
      do{
	val -= probabilities[size-minSize];
	size += (val > 0 ? 1:0);
      } while ( val > 0 );
    } while ( containsEnough(size) );
    /* size = size of the next subtable we are going to use */
    
    IntegerSet instSet = new IntegerSet(size);
    do{
      instSet.clear();
      while (instSet.size()<size) {
	Integer num = new Integer((int)(Math.random() * tableSize));
	instSet.add(num);
      } // end of while ()
    } while ( h.contains(instSet) );
    
    h.add(instSet);
    int [] indices = instSet.toIntArray();
    Instances res = new Instances(instances,indices.length);
    for (int i=0;i<indices.length;i++) {
      res.add(instances.instance(indices[i]));
    } // end of for ()
    return res;
  }

  public int numChoices(){
    return numChoices;
  }

  public static void main (String[] args) {
    try {
      double alpha = .1; 
      double maxError = .05;
      double confidence = .8;
      long start = System.currentTimeMillis();
      int size = 4;
      int numTables = 3;
      /* Do some testing */
      Reader reader = null;
      if (args.length == 0) {
	reader = new BufferedReader(new InputStreamReader(System.in));
      } else {
	reader = new BufferedReader(new FileReader(args[0]));
      }
      Instances i = new Instances(reader);
      System.out.println("Table size: " + i.numInstances());
      SubtableChooser s = new SubtableChooser(i,(int)(i.numInstances()*.7),
					      20);
      while ( s.hasNext() ) {
	System.out.println("Next chosen size: " + s.next().numInstances());
      } // end of while ()
      
      System.out.println("Done.");
    } catch (Exception e) {
      e.getMessage();
      e.printStackTrace();
    } // end of try-catch
  } // end of main ()
  

}



